@@ -27,9 +27,10 @@ The Scanner package for the SCons software construction utility.
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #
 
-__revision__ = "/home/scons/scons/branch.0/baseline/src/engine/SCons/Scanner/__init__.py 0.96.1.D001 2004/08/23 09:55:29 knight"
+__revision__ = "/home/scons/scons/branch.0/branch.96/baseline/src/engine/SCons/Scanner/__init__.py 0.96.91.D001 2005/09/08 09:14:36 knight"
 
 import re
+import string
 
 import SCons.Node.FS
 import SCons.Sig
@@ -52,22 +53,43 @@ def Scanner(function, *args, **kw):
     else:
         return apply(Base, (function,) + args, kw)
 
+
+class _Binder:
+    def __init__(self, bindval):
+        self._val = bindval
+    def __call__(self):
+        return self._val
+    def __str__(self):
+        return str(self._val)
+        #debug: return 'B<%s>'%str(self._val)
+
+BinderDict = {}
+
+def Binder(path):
+    try:
+        return BinderDict[path]
+    except KeyError:
+        b = _Binder(path)
+        BinderDict[path] = b
+        return b
+
+
 class FindPathDirs:
-    """A class to bind a specific *PATH variable name and the fs object
-    to a function that will return all of the *path directories."""
-    def __init__(self, variable, fs):
+    """A class to bind a specific *PATH variable name to a function that
+    will return all of the *path directories."""
+    def __init__(self, variable):
         self.variable = variable
-        self.fs = fs
-    def __call__(self, env, dir, argument=None):
+    def __call__(self, env, dir, target=None, source=None, argument=None):
+        # The goal is that we've made caching this unnecessary
+        # because the caching takes place at higher layers.
         try:
             path = env[self.variable]
         except KeyError:
             return ()
 
-        return tuple(self.fs.Rsearchall(env.subst_path(path),
-                                        must_exist = 0,
-                                        clazz = SCons.Node.FS.Dir,
-                                        cwd = dir))
+        path = env.subst_path(path, target=target, source=source)
+        path_tuple = tuple(env.fs.Rfindalldirs(path, dir))
+        return Binder(path_tuple)
 
 class Base:
     """
@@ -75,6 +97,9 @@ class Base:
     straightforward, single-pass scanning of a single file.
     """
 
+    if SCons.Memoize.use_memoizer:
+        __metaclass__ = SCons.Memoize.Memoized_Metaclass
+
     def __init__(self,
                  function,
                  name = "NONE",
@@ -82,7 +107,7 @@ class Base:
                  skeys = [],
                  path_function = None,
                  node_class = SCons.Node.FS.Entry,
-                 node_factory = SCons.Node.FS.default_fs.File,
+                 node_factory = None,
                  scan_check = None,
                  recursive = None):
         """
@@ -104,6 +129,8 @@ class Base:
         (a construction environment, optional directory, and optional
         argument for this instance) and returns a tuple of the
         directories that can be searched for implicit dependency files.
+        May also return a callable() which is called with no args and
+        returns the tuple (supporting Bindable class).
 
         'node_class' - the class of Nodes which this scan will return.
         If node_class is None, then this scanner will not enforce any
@@ -118,14 +145,19 @@ class Base:
         this node really needs to be scanned.
 
         'recursive' - specifies that this scanner should be invoked
-        recursively on the implicit dependencies it returns (the
-        canonical example being #include lines in C source files).
+        recursively on all of the implicit dependencies it returns
+        (the canonical example being #include lines in C source files).
+        May be a callable, which will be called to filter the list
+        of nodes found to select a subset for recursive scanning
+        (the canonical example being only recursively scanning
+        subdirectories within a directory).
 
-        The scanner function's first argument will be the name of a file
-        that should be scanned for dependencies, the second argument will
-        be an Environment object, the third argument will be the value
-        passed into 'argument', and the returned list should contain the
-        Nodes for all the direct dependencies of the file.
+        The scanner function's first argument will be a Node that should
+        be scanned for dependencies, the second argument will be an
+        Environment object, the third argument will be the tuple of paths
+        returned by the path_function, and the fourth argument will be
+        the value passed into 'argument', and the returned list should
+        contain the Nodes for all the direct dependencies of the file.
 
         Examples:
 
@@ -150,15 +182,20 @@ class Base:
         self.node_class = node_class
         self.node_factory = node_factory
         self.scan_check = scan_check
-        self.recursive = recursive
+        if callable(recursive):
+            self.recurse_nodes = recursive
+        elif recursive:
+            self.recurse_nodes = self._recurse_all_nodes
+        else:
+            self.recurse_nodes = self._recurse_no_nodes
 
-    def path(self, env, dir = None):
+    def path(self, env, dir=None, target=None, source=None):
         if not self.path_function:
             return ()
         if not self.argument is _null:
-            return self.path_function(env, dir, self.argument)
+            return self.path_function(env, dir, target, source, self.argument)
         else:
-            return self.path_function(env, dir)
+            return self.path_function(env, dir, target, source)
 
     def __call__(self, node, env, path = ()):
         """
@@ -177,18 +214,26 @@ class Base:
         kw = {}
         if hasattr(node, 'dir'):
             kw['directory'] = node.dir
+        node_factory = env.get_factory(self.node_factory)
         nodes = []
         for l in list:
             if self.node_class and not isinstance(l, self.node_class):
-                l = apply(self.node_factory, (l,), kw)
+                l = apply(node_factory, (l,), kw)
             nodes.append(l)
         return nodes
 
     def __cmp__(self, other):
-        return cmp(self.__dict__, other.__dict__)
+        try:
+            return cmp(self.__dict__, other.__dict__)
+        except AttributeError:
+            # other probably doesn't have a __dict__
+            return cmp(self.__dict__, other)
 
     def __hash__(self):
-        return hash(repr(self))
+        return id(self)
+
+    def __str__(self):
+        return self.name
 
     def add_skey(self, skey):
         """Add a skey to the list of skeys"""
@@ -202,6 +247,22 @@ class Base:
     def select(self, node):
         return self
 
+    def _recurse_all_nodes(self, nodes):
+        return nodes
+
+    def _recurse_no_nodes(self, nodes):
+        return []
+
+    recurse_nodes = _recurse_no_nodes
+
+if SCons.Memoize.use_old_memoization():
+    _Base = Base
+    class Base(SCons.Memoize.Memoizer, _Base):
+        "Cache-backed version of Scanner Base"
+        def __init__(self, *args, **kw):
+            apply(_Base.__init__, (self,)+args, kw)
+            SCons.Memoize.Memoizer.__init__(self)
+
 
 class Selector(Base):
     """
@@ -209,7 +270,7 @@ class Selector(Base):
     scanner_key() (suffix) for a specific Node.
     """
     def __init__(self, dict, *args, **kw):
-        Base.__init__(self, (None,)+args, kw)
+        apply(Base.__init__, (self, None,)+args, kw)
         self.dict = dict
 
     def __call__(self, node, env, path = ()):
@@ -246,22 +307,23 @@ class Classic(Current):
     regular expressions to find the includes.
 
     Note that in order for this to work "out of the box" (without
-    overriding the find_include() method), the regular expression passed
-    to the constructor must return the name of the include file in group
-    0.
+    overriding the find_include() and sort_key() methods), the regular
+    expression passed to the constructor must return the name of the
+    include file in group 0.
     """
 
-    def __init__(self, name, suffixes, path_variable, regex,
-                 fs=SCons.Node.FS.default_fs, *args, **kw):
+    def __init__(self, name, suffixes, path_variable, regex, *args, **kw):
 
         self.cre = re.compile(regex, re.M)
-        self.fs = fs
 
-        def _scan(node, env, path, self=self, fs=fs):
-            return self.scan(node, env, path)
+        def _scan(node, env, path=(), self=self):
+            node = node.rfile()
+            if not node.exists():
+                return []
+            return self.scan(node, path)
 
         kw['function'] = _scan
-        kw['path_function'] = FindPathDirs(path_variable, fs)
+        kw['path_function'] = FindPathDirs(path_variable)
         kw['recursive'] = 1
         kw['skeys'] = suffixes
         kw['name'] = name
@@ -269,14 +331,16 @@ class Classic(Current):
         apply(Current.__init__, (self,) + args, kw)
 
     def find_include(self, include, source_dir, path):
-        n = SCons.Node.FS.find_file(include, (source_dir,) + path, self.fs.File)
+        "__cacheable__"
+        if callable(path): path = path()
+        n = SCons.Node.FS.find_file(include, (source_dir,) + tuple(path))
         return n, include
 
-    def scan(self, node, env, path=()):
-        node = node.rfile()
+    def sort_key(self, include):
+        return SCons.Node.FS._my_normcase(include)
 
-        if not node.exists():
-            return []
+    def scan(self, node, path=()):
+        "__cacheable__"
 
         # cache the includes list in node so we only scan it once:
         if node.includes != None:
@@ -285,37 +349,27 @@ class Classic(Current):
             includes = self.cre.findall(node.get_contents())
             node.includes = includes
 
+        # This is a hand-coded DSU (decorate-sort-undecorate, or
+        # Schwartzian transform) pattern.  The sort key is the raw name
+        # of the file as specifed on the #include line (including the
+        # " or <, since that may affect what file is found), which lets
+        # us keep the sort order constant regardless of whether the file
+        # is actually found in a Repository or locally.
         nodes = []
         source_dir = node.get_dir()
         for include in includes:
             n, i = self.find_include(include, source_dir, path)
 
-            if not n is None:
-                nodes.append(n)
-            else:
+            if n is None:
                 SCons.Warnings.warn(SCons.Warnings.DependencyWarning,
                                     "No dependency generated for file: %s (included from: %s) -- file not found" % (i, node))
+            else:
+                sortkey = self.sort_key(include)
+                nodes.append((sortkey, n))
 
-        # Schwartzian transform from the Python FAQ Wizard
-        def st(List, Metric):
-            def pairing(element, M = Metric):
-                return (M(element), element)
-            def stripit(pair):
-                return pair[1]
-            paired = map(pairing, List)
-            paired.sort()
-            return map(stripit, paired)
-
-        def normalize(node):
-            # We don't want the order of includes to be
-            # modified by case changes on case insensitive OSes, so
-            # normalize the case of the filename here:
-            # (see test/win32pathmadness.py for a test of this)
-            return SCons.Node.FS._my_normcase(str(node))
-
-        transformed = st(nodes, normalize)
-        # print "Classic: " + str(node) + " => " + str(map(lambda x: str(x),list(transformed)))
-        return transformed
+        nodes.sort()
+        nodes = map(lambda pair: pair[1], nodes)
+        return nodes
 
 class ClassicCPP(Classic):
     """
@@ -328,12 +382,18 @@ class ClassicCPP(Classic):
     the contained filename in group 1.
     """
     def find_include(self, include, source_dir, path):
+        "__cacheable__"
+        if callable(path):
+            path = path()   #kwq: extend callable to find_file...
+
         if include[0] == '"':
-            n = SCons.Node.FS.find_file(include[1],
-                                        (source_dir,) + path,
-                                        self.fs.File)
+            paths = Binder( (source_dir,) + tuple(path) )
         else:
-            n = SCons.Node.FS.find_file(include[1],
-                                        path + (source_dir,),
-                                        self.fs.File)
+            paths = Binder( tuple(path) + (source_dir,) )
+
+        n = SCons.Node.FS.find_file(include[1], paths)
+
         return n, include[1]
+
+    def sort_key(self, include):
+        return SCons.Node.FS._my_normcase(string.join(include))
